#__author__ = "RVC"
#__email__= "ruben.vancoile@gmail.com"
#__date__= "2017-11-12"

from copy import deepcopy
from scipy.stats import uniform
from scipy.stats import gumbel_r
from scipy.stats import norm
from scipy.stats import lognorm
from scipy.stats import t

import numpy as np
import pandas as pd


## inverse CDF function ##
##########################

def Finv_Uniform(r,a,b):

	return uniform.ppf(r,a,b-a)

def Finv_Gumbel(r,m,s):

	scale,loc=p_Gumbel(m,s)

	return gumbel_r.ppf(r,loc,scale)

def Finv_Normal(r,m,s):

	return norm.ppf(r,m,s)

def Finv_Lognormal(r,m,s):

	sln,mln=p_Lognormal(m,s)

	return lognorm.ppf(r,sln,0,np.exp(mln))

def Finv_t(r,df):

	return t.ppf(r,df)


## CDF function ##
##################

def F_Normal(x,m,s):

	return norm.cdf(x,m,s)

def F_Lognormal(x,m,s):

	sln,mln=p_Lognormal(m,s)

	return lognorm.cdf(x,sln,0,np.exp(mln))

def F_Gumbel(r,m,s):

	scale,loc=p_Gumbel(m,s)

	return gumbel_r.cdf(r,loc,scale)


## PDF function ##
##################

def f_Normal(x,m,s):

	return norm.pdf(x,m,s)

def f_Lognormal(x,m,s):

	sln,mln=p_Lognormal(m,s)

	return lognorm.pdf(x,sln,0,np.exp(mln))


## parameter calculation ##
###########################

def p_Lognormal(m,s):

	cov=s/m;

	sln=np.sqrt(np.log(1+cov**2))
	mln=np.log(m)-1/2*sln**2

	return sln,mln

def p_Gumbel(m,s):

	# parameters Gumbel W&S
	alpha=1.282/s
	u=m-0.5772/alpha

	# parameters Gumbel scipy
	scale=1/alpha
	loc=u

	return scale,loc

## 'moment' calculation ##
##########################

def m_Lognormal(mln,sln):

	cov=np.sqrt(np.exp(sln**2)-1)
	m=np.exp(mln+1/2*sln**2)

	return m,m*cov

## post-processing ##
#####################

def MCStoPDF(sMC,binN):
	# USE: transforms pd_Series into PDF_histogram pd_DataFrame

	# histogram calculation
	(hist,binEdge)=np.histogram(sMC,bins=binN)
	# print(hist)
	# print(binEdge)

	# relative frequency and PDF scaling
	nMCS=len(sMC.index)
	delta_bin=binEdge[2]-binEdge[1]

	relFreq=hist/nMCS
	PDF=relFreq/delta_bin # integral over bin-width equal to relFreq

	# mid-bin listing
	x=(binEdge[1:]+binEdge[:-1])/2

	# pandas DataFrame output
	out=pd.DataFrame([x,PDF],index=['X','mcsPDF'])


	return out.transpose()


#__author__ = "RVC"
#__email__= "ruben.vancoile@gmail.com"
#__date__= "2017-11-24"

import pandas as pd
import sympy as sy
import numpy as np

import statFunc
from PrintAuxiliary import Print_DataFrame


########################
## USE AND BACKGROUND ##
########################
#
#
#

###############################
## (TEMPORARY LOCAL SUPPORT) ##
###############################

def ParameterRealization_r(varDict,rArray):
	# evaluates parameter realizations
	# input
	# * parameterDict with info stochastic variable
	# * array of r-values

	## List of modifications
	# 2017-11-24: created

	DistType=varDict['Dist']

	if DistType=='N':
		return statFunc.Finv_Normal(rArray,varDict['m'],varDict['s'])
	if DistType=='LN':
		return statFunc.Finv_Lognormal(rArray,varDict['m'],varDict['s'])
	if DistType=='G':
		return statFunc.Finv_Gumbel(rArray,varDict['m'],varDict['s'])
	if DistType=='DET':
		return np.ones(np.shape(rArray))*varDict['m']

def LS_evaluation(LS_eval,nvar,xMatrix):
	# evaluates limitstate (more generally: evaluates function) 
	# input
	# * lambdified function
	# * number of variables
	# * matrix of input values, with row for each set of parameter realizations
	# note
	# ordering of matrix input values must correspond with lambdify order

	## List of modifications
	# 2017-12-24: created

	#
	### CODING TO BE IMPROVED ###
	#

	# currently hardcoded number of subs
	# more Pythonic code? !?!?!?
	if nvar==1:
		lsEvalList=LS_eval(xMatrix[:,0])
	elif nvar==2:
		lsEvalList=LS_eval(xMatrix[:,0],xMatrix[:,1])
	elif nvar==3:
		lsEvalList=LS_eval(xMatrix[:,0],xMatrix[:,1],xMatrix[:,2])
	elif nvar==4:
		lsEvalList=LS_eval(xMatrix[:,0],xMatrix[:,1],xMatrix[:,2],xMatrix[:,3])
	elif nvar==5:
		lsEvalList=LS_eval(xMatrix[:,0],xMatrix[:,1],xMatrix[:,2],xMatrix[:,3],xMatrix[:,4])
	elif nvar==6:
		lsEvalList=LS_eval(xMatrix[:,0],xMatrix[:,1],xMatrix[:,2],xMatrix[:,3],xMatrix[:,4],xMatrix[:,5])
	## to be continued

	return lsEvalList



###############
## FUNCTIONS ##
###############

def MonteCarlo(limitstate,ParameterDict,nMC):
	# performs crude Monte Carlo simulation
	# input
	# * limitstate: symbolic limit state function
	# * ParameterDict: dictionary of all parameters, including probabilistic discription
	# * nMC: number of simulations
	# output
	# * array of limit state evaluations, DataFrame with parameter values and limit state valuation, DataFrame with random values

	## List of modifications
	# 2017-11-24: created
	# 2017-12-24: LS_evaluation outside of MonteCarlo

	## create MCS array random values
	symbolList=limitstate.atoms(sy.Symbol)
	nvar=len(symbolList)
	rMatrix=np.random.rand(nMC,nvar)

	## calculate MCS array random realizations
	xMatrix=np.zeros((nMC,nvar))
	# indexing in the ParameterDict - general indexing  
	indexDict={}
	for key in ParameterDict:
		indexDict[ParameterDict[key]['name']]=key
	# random realization per parameter
	varOrder=[]
	for i,X in enumerate(symbolList):
		# save order for printing lists
		var=X.name
		varOrder.append(var)
		# collect Dict (stochastic) variable
		localDict=ParameterDict[indexDict[var]]
		xMatrix[:,i]=ParameterRealization_r(localDict,rMatrix[:,i])

	## evaluate limit state
	limitstateEval = sy.lambdify(tuple(varOrder), limitstate, 'numpy')
	lsEvalList=LS_evaluation(limitstateEval,nvar,xMatrix)


	## output
	# outR first, before appending the varOrder...
	outR=pd.DataFrame(rMatrix,columns=varOrder)
	# set output - variable realizations + limit state evaluation
	lsEvalList=np.reshape(lsEvalList,(nMC,1))
	full=np.concatenate((xMatrix,lsEvalList),axis=1)
	varOrder.append('LS')
	outX=pd.DataFrame(full,columns=varOrder)

	return lsEvalList, outX, outR

def Taylor(limitstate,ParameterDict):
	# Taylor approximation of mean value and standard deviation
	# input
	# * limitstate: symbolic limit state function
	# * ParameterDict: dictionary of all parameters, including probabilistic discription
	# output
	# * mean value, standard deviation

	# 2017-11-24: created

	# list of symbols in limitstate
	symbolList=limitstate.atoms(sy.Symbol)
	nvar=len(symbolList)
	xMatrix=np.zeros((2,nvar)) 	# array mean values and standard deviations

	# indexing in the ParameterDict - general indexing  
	indexDict={}
	for key in ParameterDict:
		indexDict[ParameterDict[key]['name']]=key

	# values per parameter
	varOrder=[]
	lsList=[]
	for i,X in enumerate(symbolList):
		# save order for printing lists
		var=X.name
		varOrder.append(var)
		# collect Dict (stochastic) variable
		localDict=ParameterDict[indexDict[var]]
		xMatrix[:,i]=[localDict['m'],localDict['s']]
		# take derivative of limitstate function
		ls=sy.diff(limitstate,X)
		lsList.append(ls)

	## evaluate limit state - mean value
	limitstateEval = sy.lambdify(tuple(varOrder), limitstate, 'numpy')	
	m=LS_evaluation(limitstateEval,nvar,xMatrix); m=m[0]

	s2=0
	for i in np.arange(nvar):
		functionEval = sy.lambdify(tuple(varOrder), lsList[i], 'numpy')	
		d=LS_evaluation(functionEval,nvar,xMatrix)
		if type(d)!=int: d=d[0]
		s2=s2+d**2*xMatrix[1,i]**2

	return m,np.sqrt(s2)



##########
## TEST ##
##########

# # limit state
# r,e1,e2=sy.symbols("r e1 e2")
# g=r-e1*e2

# # number of Monte Carlo
# nsim=10

# # ParameterDict
# mr=20; sr=3
# me1=3; se1=2
# me2=5; se2=3

# R={
# 'name':'r',
# 'Dist':'N',
# 'DIM':"[N]",
# 'm':mr,
# 's':sr,
# 'info':''
# }

# E1={
# 'name':'e1',
# 'Dist':'LN',
# 'DIM':"[N]",
# 'm':me1,
# 's':se1,
# 'info':''
# }

# E2={
# 'name':'e2',
# 'Dist':'G',
# 'DIM':"[N]",
# 'm':me2,
# 's':se2,
# 'info':''
# }

# Dict={
# 	'01':R,
# 	'02':E1,
# 	'03':E2
# 	}

# # Z,dfX,dfr=MonteCarlo(g,Dict,nsim)
# # Print_DataFrame([dfr,dfX],'TestOutput/test3var',['r','X'])
# # print(dfr)

# [m,s]=Taylor(g,Dict)


# print(m,s)


# ------------------------------------------------------------------------------------------------------------
# STANDARDIZED VISUALIZATION MAGNELPY
# 	standardized visualization (colours, lines, labels...)
# 	for standardized input (pd.DataFrame...)
#
# Notes
#	* syntax sub- and superscripts: 'Xijk' write as '$X_{ijk}$''

# Wouter Botte, Ruben Van Coile - 2018
# cfr. Wouter Botte - Figures.py - PhD dissertation 2017
# ------------------------------------------------------------------------------------------------------------

####################
## MODULE IMPORTS ##
####################

import matplotlib.pyplot as plt
import numpy as np
import sys

####################
## MODULE IMPORTS ##
####################

font = {'size' : 8}
plt.rc('font', **font)

###################
## AUX FUNCTIONS ##
###################

# def LinePlot_XmultiY(df,saveDir=''):

def LinePlot_XmultiY(df,axisLabels=['X','Y'],Xticks=None,Yticks=None,SW_show=False,savePath=None,
	SW_return=False,Xminorticks=None,Yminorticks=None,SW_grid=False):
	## lineplot visualization XmultiY
	# note: additional arguments could be packed in a DataFrame with columns X and Y, and different indices
	#
	#
	## input explanation
	# df 			pd.DataFrame with index values X-axis and columns as Y-valuex
	#				column labels used as legend labels
	# Xticks 		np.array with major X ticks
	# Yticks 		np.array with major Y ticks
	# SW_show		boolean indicating immediate visualization of plot yes/no
	# savePath		path to save figure. Figure not saved if no path provided
	# SW_return		boolean. Returns the figure object (for further custom manipulation) if True
	# Xminorticks	np.array with minor X ticks
	# Yminorticks	np.array with minor Y ticks
	# SW_grid		boolean for gridlines
	#

	## unpack
	# X values and bounds
	X=df.index.values

	## initialization of plot - hardcoded parameters magnelpy
	fig, ax1 = plt.subplots(figsize=(3.5,2.77))
	colors = ['k', 'r', 'green', 'orange', 'blue', 'm', 'c', 'grey', 'brown']
	linestyles = ['-', ':', '--','-.','--']
	dash4=[20,10]

	## generation of plot from df data
	for i,key in enumerate(df.columns):
		tmp,=ax1.plot(X, df[key], color=colors[i], label = key, lw=0.5, linestyle=linestyles[i])
		if i==4: tmp.set_dashes(dash4)

	## plot formatting
	# set size
	ax1.legend(ncol=1,fontsize=7.5)
	# set ticks
	if Xticks is not None: ax1.set_xlim(Xticks[0], Xticks[-1]); plt.xticks(Xticks)
	if Yticks is not None: ax1.set_ylim(Yticks[0], Yticks[-1]); ax1.set_yticks(Yticks)
	ax1.minorticks_on()
	if Xminorticks is not None: ax1.set_xticks(Xminorticks,minor=True)
	if Yminorticks is not None: ax1.set_yticks(Yminorticks,minor=True)
	# set labels
	ax1.set_xlabel(axisLabels[0])
	ax1.set_ylabel(axisLabels[1])
	# set grids
	if SW_grid: ax1.grid(which='both',axis='both',color='gray', linestyle='-', linewidth=0.3)

	## plot visualization and printing
	fig.tight_layout()
	if savePath is not None:
		plt.savefig(savePath, dpi=300)
	if SW_show: plt.show()

	if SW_return:
		return fig,ax1	

def histogram(df,Xticks=None,Yticks=None,width=None,Xlabel='X',Ylabel='PDF [-]'):
	# bar plot of pd.DataFrame, index as X, column as Y
	# not finalized
	# To Do: automate ticks

    # this is for plotting purpose
    index = df['X'].values
    height = df['mcsPDF'].values

    # Xticks and Yticks
    if Xticks is None:
    	xmin=min(index); xminO=np.floor(np.log10(xmin)); Xmin=10**xminO; Xmin*=np.floor(xmin/Xmin)
    	# xmax=max(index); xmaxO=np.floor(np.log10(xmax)); Xmax=10**xmaxO; Xmax*=np.ceil(xmax/Xmax)
    	xmax=max(index); Xmax=10**xminO; Xmax*=(np.ceil(xmax/Xmax)+1)
    	Xticks=np.arange(Xmin,Xmax,10**xminO)
    if Yticks is None:
    	ymax=max(height); ymaxO=np.floor(np.log10(ymax)); Ymax=10**ymaxO; Ymax*=np.ceil(ymax/Ymax)
    	Yticks=np.linspace(0,Ymax,10)
    # width of bars
    if width is None:
    	width=index[1]-index[0]
    	print(width)

    fig= plt.bar(index,height,width)
    plt.xlabel(Xlabel, fontsize=10)
    plt.ylabel(Ylabel, fontsize=10)
    plt.xticks(Xticks, fontsize=8)
    plt.yticks(Yticks,fontsize=8)
    plt.title('Observed histogram')
    plt.show()


#__author__ = "RVC"
#__email__= "ruben.vancoile@gmail.com"
#__date__= "2017-11-10"

import numpy as np
import pandas as pd
from PrintAuxiliary import *
from scipy.stats import *

################################
### Latin Hypercube function ###
################################

def LHS_rand(number_sim,number_variables,method='Olsson',test=False):
    ## sample (r-values) generation according to LHS scheme
    # method: r-value selection in specified interval
        # 'Olsson': random selection
        # 'Center': center of interval selection

    ### initialise ###
    ##################

    P=np.zeros((number_sim,number_variables))    

    ### control variables ###
    #########################

    ### basic setup of integers and random permutation ###
    ######################################################

    x0=np.arange(1,number_sim+1,1)

    for j in np.arange(number_variables):

        P[:,j]=np.random.permutation(x0) 

    P=pd.DataFrame(P)


    ### correction for correlation ###
    ##################################

    Yx=P.div(number_sim+1)
    Y=Yx.apply(norm.ppf)

    Y=Y.as_matrix()
    COV_Y=np.cov(Y,rowvar=0)
    L=np.linalg.cholesky(COV_Y)

    Y=np.dot(Y,np.transpose(np.linalg.inv(L)))
    #Yx=pd.DataFrame(Y)

    ### updated P-matrix (ranks of Y) ###
    #####################################

    Y=pd.DataFrame(Y)
    Y2=Y.rank()

    ### sampling matrix via ranks & aselect ###
    ###########################################
    if method == 'Olsson':
        R=pd.DataFrame(np.random.rand(number_sim,number_variables))
    elif method == 'Center':
        R=pd.DataFrame(0.5*np.ones((number_sim,number_variables)))
    Y3=Y2.subtract(R)

    S=Y3.div(number_sim)

    ### output ###
    ##############

    if test:

        ### printing for testing ###
        Sx=P.subtract(R)
        Sx=Sx.div(number_sim)
        COV_Sx=pd.DataFrame(np.corrcoef(Sx,rowvar=0))
        L=pd.DataFrame(L)
        COV_S=pd.DataFrame(np.corrcoef(S,rowvar=0))

        return P, Sx, COV_Sx, L, Y2, S, COV_S

    else:

        ### standard output ###
        return S





#################
### TEST-ZONE ###
#################

if __name__ == "__main__": 
    
    number_sim=100
    number_variables=5

    # LHS_rand(number_sim,number_variables)

    P, Sx, COV_Sx, L, Y2, S, COV_S=LHS_rand(number_sim,number_variables,'Olsson',True)
    # P, Sx, COV_Sx, L, Y2, S, COV_S=LHS_rand(number_sim,number_variables,'Center',True)


    Print_DataFrame([P, Sx, COV_Sx, L, Y2, S, COV_S],'LHS_TMP',['P', 'Sx', 'COV_Sx', 'L', 'Y2', 'S', 'COV_S'])


#__author__ = "RVC"
#__email__= "ruben.vancoile@gmail.com"
#__date__= "2017-11-23"

import numpy as np


## EN 1990 based functions ##
#############################

def Gk_ULS(Rd,chi,u,ULS):
	# calculate MGk for bending ultimate limit state
	# for given load ratio chi = MQk/(MQk+MGk)
	# for given utilisation u

	# parameters
	gamma_G=1.35
	psi_0=0.7
	gamma_Q=1.5
	dzeta=0.85

	if ULS=='STR':
		return u*Rd/np.maximum(gamma_G+psi_0*gamma_Q*chi/(1-chi),dzeta*gamma_G+gamma_Q*chi/(1-chi))
	elif ULS=='EQU':
		return u*Rd/(gamma_G+gamma_Q*chi/(1-chi))

def kfi_req(chi,u,gR=1,ULS='STR'):
	# calculate utilisation fire (time 0 - i.e. no temperature effect), considering utilisation at ambient u
	# for given load ratio chi = Qk/(Qk+Gk)
	#
	# ULS: Rd=Rk/gR
	# ULS: Ed=u*Rd=gE*Ek
	# fire: Rd,fi=kfi*Rk
	# fire: Edfi=Ek=Ed/gE=u*Rd/gE=u*Rk/gE/gR=u*Rd,fi/kfi/gE/gR=ufi*Rdfi/kfi
	# ==> ufi=u/gE/gR

	## history log ##
	# 2017-12-15: created

	# parameters
	gamma_G=1.35
	psi_0=0.7
	gamma_Q=1.5
	dzeta=0.85

	psi_fi=0.5

	if ULS=='STR':
		return u*(1+psi_fi*chi/(1-chi))/np.maximum(gamma_G+psi_0*gamma_Q*chi/(1-chi),dzeta*gamma_G+gamma_Q*chi/(1-chi))
	elif ULS=='EQU':
		return u*(1+psi_fi*chi/(1-chi))/(gamma_G+gamma_Q*chi/(1-chi))


#__author__ = "RVC"
#__email__= "ruben.vancoile@gmail.com"
#__date__= "2015-07-19"

import numpy as np
import pandas as pd
import os, shutil




######################## 
###  READ FUNCTIONS  ###
########################


def Read_DataFrame(path,sheetList):
	# read multiple sheets from *.xlsx worksheet

	xl = pd.ExcelFile(path) # *.xlsx with input values

	out={} # initialise output

	for i,sheet in enumerate(sheetList):

		# read *.xlsx worksheeet as df
		local=xl.parse(sheet)
		# assign to Dict
		out[sheet]=local
		
	return out


######################### 
###  PRINT FUNCTIONS  ###
#########################

def Print_DataFrame(mytotallist,name,sheet):

	writer=pd.ExcelWriter(name+'.xlsx')
	for n,df in enumerate(mytotallist):
		Single_Print_DataFrame(df,writer,sheet[n])
	writer.save()
	writer.close()

def Single_Print_DataFrame(df,writer,sheet):
	df.to_excel(writer,sheet)

def RemoveFolderData(folder):
	for the_file in os.listdir(folder):
		file_path=os.path.join(folder, the_file)
		if os.path.isfile(file_path):
			os.unlink(file_path)

def PrintMaxEnt(data,targetdir,targetfilename,method='MCS'):

	if method=='MCS':
		# data as pd.Series
		n=data.index.size;
		index=np.arange(1,n+1)
		df=pd.DataFrame(data,columns=[data.name]); df.index=index
		Print_DataFrame([df],targetdir+'/'+targetfilename,['DATA'])


############################## 
###  MAIN CONTROL - DEBUG  ###
##############################


if __name__ == "__main__":


	## df path with out results ##
	outpath="C:\\Users\\rvcoile\\OneDrive - UGent\\CalcCenter\\Results\\ISOcc120min\\outputFull.xlsx"
	sheet1='out'
	corrsheet='corrList1'

	## trial function ##
	read=Read_DatFrame(outpath,[sheet1,corrsheet])

	## print output ##
	print(read[corrsheet])




#__author__ = "RVC"
#__email__= "ruben.vancoile@gmail.com"
#__date__= "2017-12-25"

import numpy as np


## auxiliary function for common use ##
#######################################

# def np_list_tile(list,m,n):
# 	# input:
# 	# * list to be tiled
# 	# * m: row repeats tiling
# 	# * n: column repeats tiling

# 	U=np.array(list).reshape(1,nu)
# 	return np.tile(U,(nchi,1))


# __author__ = "RVC"
# __email__= "ruben.vancoile@gmail.com"
# __date__= "2018-06-28"

#
# sampling scheme for external model input
# L=5 is hardcoded - limited variable may exist at different locations, but has not been finalized
#


####################
## MODULE IMPORTS ##
####################

import pandas as pd
import numpy as np
import sys


directory="C:/Users/rvcoile/Google Drive/Research/Codes/Python3.6/REF/rvcpy"
sys.path.append(directory)
from PrintAuxiliary import Print_DataFrame

# from statFunc import F_Normal

###############
## HARDCODED ##
###############

##############
## FUNCTION ##
##############

##########
## CORE ##
##########

## user input ##
################

# filename
filename=input("\nPlease provide path to input file Gauss samples (*.xlsx): ")
if filename[0]=="\"": filename=filename[1:-1] # strips quotes from path

# sheet Excel file with data
print("\n## Gauss samples in sheet 'modelInput'. Note required layout. ##")
u=input("Press ENTER to confirm, or provide name of worksheet: ")
if u!='': sheet=u;
else: sheet='DATA'
print("\nWorksheet set to ", sheet)

# Test to be performed
print("Please provide number of Test to be performed: ")
print("1 : direct test single variable")
u=input("\n")
test=int(u)

## Test application ##
######################

if test==1:
    Y=Test01(X)


## MaxEnt output generation ##
##############################

# Print_DataFrame([samples_modelInput,samplingScheme],outdir+'/MDRMGauss_samples',['modelInput','samplingScheme'])

##########
## TEST ##
##########

print(test)


#__author__ = "RVC"
#__email__= "ruben.vancoile@gmail.com"
#__date__= "2017-12-14"


import numpy as np

from statFunc import Finv_Normal


## structural steel reduction function ##
#########################################

def kfy_struct(Theta,method='Negar',r=0):

	if method=='Negar':
		return kfy_struct_Negar(Theta,r,1)

	return 0

def kfy_struct_Negar(Theta,r,base=1,methodEN=2):
	# (Negar Elhami Khorasani, 2015) as given in nota Thomas Gernay: MaxEnt-1panel_V2.docx

	if base==1: # no base
		# regression coefficients (Khorasani, 2015), p.128, Eq. 5.6, as given in Excel datasheet
		t1=1.610513417
		t2=-0.001675254547
		t3=-0.000003363505313
		t4=0.3543142209

		# error term regression
		eps=Finv_Normal(r,0,1)

		# evaluation
		Z=t1+t2*Theta+t3*Theta**2+t4*eps

		return 1.2*np.exp(Z)/(np.exp(Z)+1) # kfy, Eq. 5.6 (Khorasani, 2015)
	elif base==2: # NIST base
		return 0
	elif base==3: # EC base
		# currently found to be problematic

		kfyEN=kfy_struct_EN(Theta,methodEN) # [-] Eurocode nominal reduction factor
		# method 2 ==> utility formula section 4.2.4 as default EN 1993-1-2

		# base value
		rlogit=np.log(((kfyEN+10**-6)/1.7)/(1-(kfyEN+10**-6)/1.7))
		eps=Finv_Normal(r,0,1)

		# evaluation
		teller=1.7*np.exp(rlogit+0.412-0.81*10**-3*Theta+0.58*10**-6*Theta**1.9+0.43*eps)
		noemer=(1+np.exp(rlogit+0.412-0.81*10**-3*Theta+0.58*10**-6*Theta**1.9+0.43*eps))
		kfy=teller/noemer

	return kfy

def kfy_struct_EN(Theta,method=2):

	if method==1:
		# tabulated data, Table 3.1, EN 1993-1-2

		# not programmed
		return 0
	elif method==2:
		# utilisation formula, 4.2.4, EN 1993-1-2
		return (1/(0.9674*(1+np.exp((Theta-492)/39.19))))**(1/3.833)





#__author__ = "RVC"
#__email__= "ruben.vancoile@gmail.com"
#__date__= "2017-11-24"

import pandas as pd
import sympy as sy
import numpy as np

from copy import deepcopy

import statFunc
from PrintAuxiliary import Print_DataFrame
from GaussWeightsAndPoints import GaussPoints, GaussSampleScheme


########################
## USE AND BACKGROUND ##
########################
#
#
#

###############################
## (TEMPORARY LOCAL SUPPORT) ##
###############################

def ParameterRealization_r(varDict,rArray):

	DistType=varDict['Dist']

	if DistType=='N':
		return statFunc.Finv_Normal(rArray,varDict['m'],varDict['s'])
	if DistType=='LN':
		return statFunc.Finv_Lognormal(rArray,varDict['m'],varDict['s'])
	if DistType=='G':
		return statFunc.Finv_Gumbel(rArray,varDict['m'],varDict['s'])
	if DistType=='DET':
		return np.ones(np.shape(rArray))*varDict['m']

def LS_evaluation(LS_eval,nvar,xMatrix):

	#
	### TO BE IMPROVED ###
	#

	# currently hardcoded number of subs
	# more Pythonic code? !?!?!?
	if nvar==1:
		lsEvalList=LS_eval(xMatrix[:,0])
	elif nvar==2:
		lsEvalList=LS_eval(xMatrix[:,0],xMatrix[:,1])
	elif nvar==3:
		lsEvalList=LS_eval(xMatrix[:,0],xMatrix[:,1],xMatrix[:,2])
	elif nvar==4:
		lsEvalList=LS_eval(xMatrix[:,0],xMatrix[:,1],xMatrix[:,2],xMatrix[:,3])
	elif nvar==5:
		lsEvalList=LS_eval(xMatrix[:,0],xMatrix[:,1],xMatrix[:,2],xMatrix[:,3],xMatrix[:,4])
	elif nvar==6:
		lsEvalList=LS_eval(xMatrix[:,0],xMatrix[:,1],xMatrix[:,2],xMatrix[:,3],xMatrix[:,4],xMatrix[:,5])
	## to be continued

	return lsEvalList

def MDRM_CalculationPoints(samplingScheme,GaussPoint_df):

    modelInput=pd.DataFrame(index=samplingScheme.index,columns=GaussPoint_df.columns)

    # initialize median values for all parameter realizations
    for var in modelInput.columns:
        modelInput[var]=GaussPoint_df.loc[3,var]

    # correct on every line the single modified Gauss point
    for i in modelInput.index:

        # j,l realization
        [j,l]=samplingScheme.loc[i,:]

        # assignment
        if l!=0: # 0-value corresponds with median point
            modelInput.loc[i,l]=GaussPoint_df.loc[j,l]

    return modelInput

def MDRMG_sampling(limitstate,ParameterDict,L):
	# input
	# * limitstate: symbolic limit state function - here: more appropriate function 'Y'
	# * ParameterDict: dictionary of all parameters, including probabilistic discription
	# * L: number of Gauss points per stochastic variable (default=5)
	# output
	# * Prints *.xlsx with Gauss realizations for MaxEnt2018.py application
	#
	# indexing in the ParameterDict - general indexing  
	indexDict={}
	for key in ParameterDict:
		indexDict[ParameterDict[key]['name']]=key

	## calculate Gauss point realizations
	symbolList=limitstate.atoms(sy.Symbol); n=len(symbolList) # number of stochastic variables limitstate
	nSim=(L-1)*n+1 # number of sample points - correct for odd L only... But even number inefficient in current procedure
	points=GaussPoints(L) # Gauss points for L
	r_realizations=statFunc.F_Normal(points,0,1); r_realizations=r_realizations.flatten() # quantiles corresponding with Gauss points
	
	# realizations per variable
	GaussPoint_df=pd.DataFrame(index=np.arange(1,L+1))
	varOrder=[]
	for i,var in enumerate(indexDict.keys()):
	    localDict=ParameterDict[indexDict[var]]
	    varOrder.append(var)
	    samplepoints=ParameterRealization_r(localDict,r_realizations)
	    GaussPoint_df[i+1]=samplepoints

	## assign in sampling scheme ##
	samplingScheme=GaussSampleScheme(L,n,nSim)
	samples_modelInput=MDRM_CalculationPoints(samplingScheme,GaussPoint_df); samples_modelInput.columns=varOrder

	return samples_modelInput,samplingScheme,varOrder

def VarDict_to_df(ParameterDict):
	# transforms ParameterDict syntax into pd.DataFrame syntax (MaxEnt_sampling.py)
	# but VarOrder is important

	# indexing in the ParameterDict - general indexing
	# allows to call the correct parameter-dictionary based on the name of the variable as it appears in the limitstate  
	indexDict={}
	for key in ParameterDict:
		indexDict[ParameterDict[key]['name']]=key

	print(len(indexDict.keys()))

	# define an order of variables, and assign accordingly to pd.DataFrame
	Parameter_df=pd.DataFrame(index=np.arange(len(indexDict.keys())+1),columns=['number','X','type','info1','info2','p1','p2'])
	for i,X in enumerate(indexDict.keys()):

		print(i)
		print(X)
		print()

		## WIP ##

		# save order for printing lists
		# varOrder.append(X)
		# # collect Dict (stochastic) variable
		# localDict=ParameterDict[indexDict[var]]


	# return Parameter_df,Parameter_df['X'].tolist()


###############
## FUNCTIONS ##
###############

def MonteCarlo(limitstate,ParameterDict,nMC):
	# performs crude Monte Carlo simulation
	# input
	# * limitstate: symbolic limit state function
	# * ParameterDict: dictionary of all parameters, including probabilistic discription
	# * nMC: number of simulations
	# output
	# * array of limit state evaluations, DataFrame with parameter values and limit state valuation, DataFrame with random values

	## create MCS array random values
	symbolList=limitstate.atoms(sy.Symbol)
	nvar=len(symbolList)
	rMatrix=np.random.rand(nMC,nvar)

	## calculate MCS array random realizations
	xMatrix=np.zeros((nMC,nvar))
	# indexing in the ParameterDict - general indexing  
	indexDict={}
	for key in ParameterDict:
		indexDict[ParameterDict[key]['name']]=key
	# random realization per parameter
	varOrder=[]
	for i,X in enumerate(symbolList):
		# save order for printing lists
		var=X.name
		varOrder.append(var)
		# collect Dict (stochastic) variable
		localDict=ParameterDict[indexDict[var]]
		xMatrix[:,i]=ParameterRealization_r(localDict,rMatrix[:,i])

	## evaluate limit state
	limitstateEval = sy.lambdify(tuple(varOrder), limitstate, 'numpy')
	lsEvalList=LS_evaluation(limitstateEval,nvar,xMatrix)


	## output
	# outR first, before appending the varOrder...
	outR=pd.DataFrame(rMatrix,columns=varOrder)
	# set output - variable realizations + limit state evaluation
	lsEvalList=np.reshape(lsEvalList,(nMC,1))
	full=np.concatenate((xMatrix,lsEvalList),axis=1)
	varOrder.append('LS')
	outX=pd.DataFrame(full,columns=varOrder)

	return lsEvalList, outX, outR

def Taylor(limitstate,ParameterDict):
	# Taylor approximation of mean value and standard deviation
	# input
	# * limitstate: symbolic limit state function
	# * ParameterDict: dictionary of all parameters, including probabilistic discription
	# output
	# * mean value, standard deviation

	# list of symbols in limitstate
	symbolList=limitstate.atoms(sy.Symbol)
	nvar=len(symbolList)
	xMatrix=np.zeros((2,nvar)) 	# array mean values and standard deviations

	# indexing in the ParameterDict - general indexing  
	indexDict={}
	for key in ParameterDict:
		indexDict[ParameterDict[key]['name']]=key

	# values per parameter
	varOrder=[]
	lsList=[]
	for i,X in enumerate(symbolList):
		# save order for printing lists
		var=X.name
		varOrder.append(var)
		# collect Dict (stochastic) variable
		localDict=ParameterDict[indexDict[var]]
		xMatrix[:,i]=[localDict['m'],localDict['s']]
		# take derivative of limitstate function
		ls=sy.diff(limitstate,X)
		lsList.append(ls)

	## evaluate limit state - mean value
	limitstateEval = sy.lambdify(tuple(varOrder), limitstate, 'numpy')	
	m=LS_evaluation(limitstateEval,nvar,xMatrix); m=m[0]

	s2=0
	for i in np.arange(nvar):
		functionEval = sy.lambdify(tuple(varOrder), lsList[i], 'numpy')	
		d=LS_evaluation(functionEval,nvar,xMatrix)
		if type(d)!=int: d=d[0]
		s2=s2+d**2*xMatrix[1,i]**2

	return m,np.sqrt(s2) 

def MaxEnt_GaussEval(limitstate,ParameterDict,L=5):
	# Evaluate Gauss point realizations for MaxEnt calculation
	#
	# input
	# * limitstate: symbolic limit state function - here: more appropriate function 'Y'
	# * ParameterDict: dictionary of all parameters, including probabilistic discription
	# * L: number of Gauss points per stochastic variable (default=5)
	# output
	# * Prints *.xlsx with Gauss realizations for MaxEnt2018.py application
	#


	# input evaluation
	# opportunity for future evaluation : DET-detection and omission
	symbolList=limitstate.atoms(sy.Symbol); n=len(symbolList) # number of stochastic variables limitstate

	# issue: MaxEnt sampling takes pd.DataFrame (cfr. StochVar.xlsx) as input => need to develop switch or adapt code
	# WIP - TMP hardcoded: ParameterDict as dict entry - not yet option to read *.xlsx
	# print(isinstance(ParameterDict,dict))
	# StochVar,varOrder=VarDict_to_df(ParameterDict) # transforms Dict to Stochvar-syntax

	samples_modelInput,samplingScheme,varOrder=MDRMG_sampling(limitstate,ParameterDict,L)

	# # add mean value realization if requested
	# if SW_add_mean:
	#     # initialize series
	#     s = pd.Series(index=samples_modelInput.columns); s.name='MeanEval'
	#     # iterate over stochastic variables and introduce mean value for each of them
	#     for var in StochVar.columns:
	#         local_StochVar=StochVar[var]
	#         mean=MeanEval(local_StochVar)
	#         s[local_StochVar['number']]=mean
	#     # add series to sampling dataframe
	#     samples_modelInput=samples_modelInput.append(s)

	# evaluate model Y
	Y=limitstate # for notation purposes only
	YEval = sy.lambdify(tuple(varOrder), Y, 'numpy') # note: YEval identical to limitstateEval in MCS
	MDRMG_realizations=LS_evaluation(YEval,n,samples_modelInput.values)

	# set model output conform MaxEnt input requirement
	MDRMG_out=deepcopy(samplingScheme); MDRMG_out['Y']=MDRMG_realizations
	MDRMG_out=MDRMG_out[['Y','j','l']]

	return samples_modelInput,samplingScheme,MDRMG_out

##########
## TEST ##
##########

if __name__ == "__main__": 

	## SWITCHES between tests ##
	############################

	I_test='calc'
	# 'Basis' : basic test case
	# 'SiF2018_a': SiF 2018 load testcase - KR and KE separately
	# 'calc': specific calculation
	# 'ISO_cc' : ISO cc equivalent load case

	if I_test=='Basic':
	
		# limit state
		r,e1,e2=sy.symbols("r e1 e2")
		g=r-e1*e2

		# number of Monte Carlo
		nsim=10

		# ParameterDict
		mr=20; sr=3
		me1=3; se1=2
		me2=5; se2=3

		R={
		'name':'r',
		'Dist':'N',
		'DIM':"[N]",
		'm':mr,
		's':sr,
		'info':''
		}

		E1={
		'name':'e1',
		'Dist':'LN',
		'DIM':"[N]",
		'm':me1,
		's':se1,
		'info':''
		}

		E2={
		'name':'e2',
		'Dist':'G',
		'DIM':"[N]",
		'm':me2,
		's':se2,
		'info':''
		}

		Dict={
			'01':R,
			'02':E1,
			'03':E2
			}

		# Z,dfX,dfr=MonteCarlo(g,Dict,nsim)
		# Print_DataFrame([dfr,dfX],'TestOutput/test3var',['r','X'])
		# print(dfr)

		[m,s]=Taylor(g,Dict)


		print(m,s)

	elif I_test=='SiF2018_a':
	
		# limit state
		G,Q,KR,KE=sy.symbols("G Q KR KE")
		Z=KE/KR*(G+Q)

		# parameter values
		Gk=1275 # [kN]
		Qk=425 # [kN]
		chi=0.25

		# ParameterDict
		mG=Gk; sG=0.1*mG
		mQ=0.2*Qk; sQ=1.1*mQ
		mKE=1.; sKE=0.05
		mKR=1.; sKR=0.15

		g={
		'name':'G',
		'Dist':'N',
		'DIM':"[kN]",
		'm':mG,
		's':sG,
		'info':''
		}

		q={
		'name':'Q',
		'Dist':'G',
		'DIM':"[kN]",
		'm':mQ,
		's':sQ,
		'info':''
		}

		kr={
		'name':'KR',
		'Dist':'LN',
		'DIM':"[-]",
		'm':mKR,
		's':sKR,
		'info':''
		}

		ke={
		'name':'KE',
		'Dist':'LN',
		'DIM':"[-]",
		'm':mKE,
		's':sKE,
		'info':''
		}


		Dict={
			'01':g,
			'02':q,
			'03':kr,
			'04':ke
			}

		# Z,dfX,dfr=MonteCarlo(g,Dict,nsim)
		# Print_DataFrame([dfr,dfX],'TestOutput/test3var',['r','X'])
		# print(dfr)

		[m,s]=Taylor(Z,Dict)


		print(m,s)

	elif I_test=='calc':
	
		# limit state
		G,Q,KT=sy.symbols("G Q KT")
		Z=KT*(G+Q)

		# parameter values
		Gk=2000 # [kN]
		Qk=1333 # [kN]
		chi=0.25

		# ParameterDict
		mG=Gk; sG=0.1*mG
		mQ=0.2*Qk; sQ=1.1*mQ
		# mKE=1.; sKE=0.05
		# mKR=1.; sKR=0.15
		mKT=1.0; sKT=0.05

		g={
		'name':'G',
		'Dist':'N',
		'DIM':"[kN]",
		'm':mG,
		's':sG,
		'info':''
		}

		q={
		'name':'Q',
		'Dist':'G',
		'DIM':"[kN]",
		'm':mQ,
		's':sQ,
		'info':''
		}

		kt={
		'name':'KT',
		'Dist':'LN',
		'DIM':"[-]",
		'm':mKT,
		's':sKT,
		'info':''
		}


		Dict={
			'01':g,
			'02':q,
			'03':kt,
			}

		# Z,dfX,dfr=MonteCarlo(g,Dict,nsim)
		# Print_DataFrame([dfr,dfX],'TestOutput/test3var',['r','X'])
		# print(dfr)

		[m,s]=Taylor(Z,Dict)


		print(m,s)

	elif I_test=='ISO_cc':

		# limit state
		G,Q,KR,KE=sy.symbols("G Q KR KE")
		Z=KE/KR*(G+Q)

		# parameter values
		Gk=2000. # [kN]
		Qk=1333. # [kN]

		# ParameterDict
		mG=Gk; sG=0.1*mG
		mQ=0.2*Qk; sQ=1.1*mQ
		mKE=1.; sKE=0.05
		mKR=1.; sKR=0.15

		g={
		'name':'G',
		'Dist':'N',
		'DIM':"[kN]",
		'm':mG,
		's':sG,
		'info':''
		}

		q={
		'name':'Q',
		'Dist':'G',
		'DIM':"[kN]",
		'm':mQ,
		's':sQ,
		'info':''
		}

		kr={
		'name':'KR',
		'Dist':'LN',
		'DIM':"[-]",
		'm':mKR,
		's':sKR,
		'info':''
		}

		ke={
		'name':'KE',
		'Dist':'LN',
		'DIM':"[-]",
		'm':mKE,
		's':sKE,
		'info':''
		}


		Dict={
			'01':g,
			'02':q,
			'03':kr,
			'04':ke
			}

		[m,s]=Taylor(Z,Dict)


		print(m,s)



#__author__ = "RVC"
#__email__= "ruben.vancoile@gmail.com"
#__date__= "2017-11-10"

####################
## MODULE IMPORTS ##
####################

# general modules Python
import numpy as np
import pandas as pd
import matplotlib.pyplot as plt

# local modules rvcpy
import k_reduxTheta as kT
from statFunc import F_Normal
from PrintAuxiliary import Print_DataFrame
from EN1990 import kfi_req

from visualization import LinePlot_XmultiY

################
## Test Cener ##
################


baseMethod_Negar=1
# 1: noBase
# 2: NIST - not programmed
# 3: EC3


# temperatures
T=np.arange(0,1200+1,10)

# probabilities stdLow, mean, stdHigh
rlow=F_Normal(-1,0,1)
rmid=F_Normal(0,0,1)
rhigh=F_Normal(1,0,1)

# corresponding kfy 
kfyLow=kT.kfy_struct_Negar(T,rlow,baseMethod_Negar)
kfyMid=kT.kfy_struct_Negar(T,rmid,baseMethod_Negar)
kfyHigh=kT.kfy_struct_Negar(T,rhigh,baseMethod_Negar)

# dataframe
df_kfy=pd.DataFrame([kfyLow,kfyMid,kfyHigh],index=['$k_{fy,Low}$','kfyMid','kfyHigh'],columns=T)
df_kfy=df_kfy.transpose()

# visualize
axisLabels=['T [°C]','kfy [-]']
Ybounds=[0,1.2]
Xticks=np.arange(0,1200+1,200)
Yticks=np.arange(0,1.2+0.1,0.2)
savePath='test.png'
LinePlot_XmultiY(df_kfy,axisLabels=axisLabels,Xticks=Xticks,Yticks=Yticks,savePath=savePath)
# LinePlot_XmultiY(df_kfy,axisLabels=axisLabels)

# plt.figure('kfy')
# plt.plot(df_kfy.index,df_kfy.loc[:,['kfyLow','kfyMid','kfyHigh']])
# plt.grid(True)
# plt.xlabel('Theta [C]')
# plt.ylabel('kfy [-]')
# plt.show()

